(function ($, Drupal, drupalSettings, Sortable) {
  Drupal.behaviors.entityRefTagify = {
    attach: function (context) {
      $(once('tagify-widget', 'input.ref-tagify-widget', context)).each(function () {
        const input = this;
        const { identifier, cardinality } = input.dataset;
        let controller;

        // Helper functions
        const helpers = {
          countSelectedTags() {
            return $(`.${identifier} .tagify__tag`).length;
          },

          isTagLimitReached() {
            return cardinality > 0 && this.countSelectedTags() >= cardinality;
          },

          toggleLoadingText(show = true) {
            const $tagsElement = $(`.${identifier}`);
            const $loadingText = $tagsElement.find('.tagify--loading-text');

            if (show && !$loadingText.length) {
              $('<div class="tagify--loading-text">Loading...</div>').appendTo($tagsElement);
            } else if (!show && $loadingText.length) {
              $loadingText.remove();
            }
          },

          isValidImageSource(str) {
            if (!str) return false;
            const pattern = new RegExp(
              '^(https?:\\/\\/)?' +
              '((([a-z\\d]([a-z\\d-]*[a-z\\d])*)\\.)+[a-z]{2,}|' +
              '((\\d{1,3}\\.){3}\\d{1,3}))' +
              '(\\:\\d+)?' +
              '(\\/[-a-z\\d%_.~+]*)*' +
              '(\\.(?:jpg|jpeg|png|gif|bmp|svg|webp))' +
              '(\\?[;&a-z\\d%_.~+=-]*)?' +
              '(\\#[-a-z\\d_]*)?$',
              'i'
            );
            return pattern.test(str);
          },

          highlightMatchingLetters(inputTerm, searchTerm) {
            if (!searchTerm) return inputTerm;

            const escapedSearchTerm = searchTerm.replace(/[.*+?^${}()|[\]\\]/g, '\\$&');
            const regex = new RegExp(`(${escapedSearchTerm})`, 'gi');
            return inputTerm.replace(regex, '<strong>$1</strong>');
          },

          containsHTML(str, selector) {
            if (!str || typeof str !== 'string') return false;
            return $('<div>').html(str).find(selector).length > 0;
          },

          entityIdMarkup(entityId) {
            if (!parseInt(input.dataset.showEntityId, 10) || !entityId) return '';

            return $('<div>', {
              id: 'tagify__tag-items',
              class: 'tagify__tag_with-entity-id'
            }).append(
              $('<div>', { class: 'tagify__tag__entity-id-wrap' }).append(
                $('<span>', { class: 'tagify__tag-entity-id', text: entityId })
              )
            ).prop('outerHTML');
          },

          infoLabelMarkup(infoLabel) {
            if (!infoLabel) return '';

            const $wrapper = $('<div>', {
              id: 'tagify__tag__info-label-wrap',
              class: 'tagify__tag__info-label-wrap'
            });

            if (this.isValidImageSource(infoLabel) ||
              this.containsHTML(infoLabel, 'svg') ||
              this.containsHTML(infoLabel, 'img')) {
              const $imageDiv = $('<div>', { class: 'tagify__tag-info-label-image' });

              if (this.isValidImageSource(infoLabel)) {
                $imageDiv.append(
                  $('<img>', {
                    src: infoLabel,
                    onerror: "this.style.visibility='hidden'"
                  })
                );
              } else {
                $imageDiv.html(infoLabel);
              }

              return $wrapper.append($imageDiv).prop('outerHTML');
            }

            return $wrapper.append(
              $('<span>', { class: 'tagify__tag-info-label', html: infoLabel })
            ).prop('outerHTML');
          }
        };

        // Template functions
        const templates = {
          tagMarkup(tagLabel, tagInfoLabel, tagEntityId) {
            const textClass = tagEntityId ? 'tagify__tag-text-with-entity-id' : 'tagify__tag-text';
            return `<div id="tagify__tag-items">
                     ${tagEntityId}
                     <span class="${textClass}">${tagLabel}</span>
                     ${tagInfoLabel}
                   </div>`;
          },

          tagTemplate(tagData) {
            const label = tagData.label ?? tagData.value;
            const infoLabel = helpers.infoLabelMarkup.call(helpers, tagData.info_label);
            const entityId = helpers.entityIdMarkup.call(helpers, tagData.entity_id);

            return `<tag title="${tagData.label}"
                     contenteditable='false'
                     spellcheck='false'
                     tabIndex="-1"
                     class="tagify__tag ${tagData.class || ''}"
                     ${this.getAttributes(tagData)}>
                     <x id="tagify__tag-remove-button"
                       title='Remove ${tagData.label}'
                       class='tagify__tag__removeBtn'
                       role='button'
                       aria-label='remove ${tagData.label} tag'
                       tabindex="0">
                     </x>
                     ${this.tagMarkup(label, infoLabel, entityId)}
                  </tag>`;
          },

          suggestionItemTemplate(tagData) {
            if (helpers.isTagLimitReached()) return '';
            const { class: tagClass = '', label, info_label } = tagData;
            const attributes = this.getAttributes(tagData);
            const highlightedLabel = helpers.highlightMatchingLetters(label, this.state?.inputText ?? '');
            const infoLabel = helpers.infoLabelMarkup(info_label);
            return `<div ${attributes} class="tagify__dropdown__item ${tagClass}" tabindex="0" role="option">
            <div class="tagify__dropdown__item-highlighted">${highlightedLabel}</div>
            ${infoLabel}
          </div>`;
          },

          suggestionFooterTemplate() {
            if (!helpers.isTagLimitReached()) return '';

            const limitMsg = drupalSettings.tagify.information_message.limit_tag;
            return $('<footer>', {
              'data-selector': 'tagify-suggestions-footer',
              class: this.settings.classNames.dropdownFooter
            }).append(
              $('<p>').html(`${limitMsg} <strong>${cardinality}</strong>`)
            ).prop('outerHTML');
          },

          dropdownItemNoMatch: (data) => {
            if (helpers.isTagLimitReached()) return '';

            const noMatchMsg = drupalSettings.tagify.information_message.no_matching_suggestions;
            return $('<div>', {
              class: `${tagify.settings.classNames.dropdownItem} tagify--dropdown-item-no-match`,
              value: 'noMatch',
              tabindex: '0',
              role: 'option'
            }).append(
              $('<p>').text(noMatchMsg),
              $('<strong>', { class: 'tagify--value', text: data.value })
            ).prop('outerHTML');
          }
        };

        // Core functionality
        const handleAutocomplete = Drupal.debounce((value, selectedEntities) => {
          if (controller) {
            controller.abort();
          }

          controller = new AbortController();

          if (identifier) {
            helpers.toggleLoadingText(true);
          }

          tagify.loading(value !== '');

          // Prepare URL for fetch
          const autocompleteUrl = new URL(
            $(input).data('autocomplete-path'),
            window.location.origin
          );
          const operator = autocompleteUrl.search ? '&' : '?';
          const url = `${$(input).data('autocomplete-path')}${operator}q=${encodeURIComponent(value)}&selected=${selectedEntities}`;

          // Fetch suggestions
          $.ajax({
            url: url,
            dataType: 'json',
            beforeSend: function() {
              if (controller) controller.signal;
            }
          })
            .done(function(newWhitelist) {
              const newWhitelistData = $.map(newWhitelist, function(current) {
                return {
                  value: current.entity_id,
                  entity_id: current.entity_id,
                  info_label: current.info_label,
                  label: current.label,
                  editable: current.editable,
                  input: tagify.state.inputText,
                  ...current.attributes
                };
              });

              tagify.whitelist = newWhitelistData;

              if (identifier) {
                helpers.toggleLoadingText(false);
              }

              tagify.loading(false).dropdown.show(value);
            })
            .fail((error) => {
              if (!(error instanceof Error && error.name === 'AbortError')) {
                console.error('Error fetching data:', error);
              }
            });
        }, 250);

        // Initialize Tagify
        const tagify = new Tagify(input, {
          dropdown: {
            enabled: parseInt(input.dataset.suggestionsDropdown, 10),
            highlightFirst: true,
            fuzzySearch: !!parseInt(input.dataset.matchOperator, 10),
            maxItems: input.dataset.maxItems ?? Infinity,
            closeOnSelect: true,
            searchKeys: ['label', 'input'],
            mapValueTo: 'label',
          },
          templates: {
            tag: templates.tagTemplate.bind({ ...templates, getAttributes: (...args) => tagify.getAttributes(...args) }),
            dropdownItem: templates.suggestionItemTemplate.bind({ ...templates, getAttributes: (...args) => tagify.getAttributes(...args) }),
            dropdownFooter: templates.suggestionFooterTemplate.bind({ ...templates }),
            dropdownItemNoMatch: Drupal.debounce(templates.dropdownItemNoMatch, 250),
          },
          whitelist: [],
          placeholder: parseInt(input.dataset.placeholder, 10),
          tagTextProp: 'label',
          editTags: false,
          maxTags: cardinality > 0 ? cardinality : Infinity,
        });

        // Configure tag enforcement settings
        tagify.settings.enforceWhitelist = helpers.isTagLimitReached() && cardinality > 1
          ? false
          : !$(input).hasClass('tagify--autocreate');

        tagify.settings.skipInvalid = helpers.isTagLimitReached()
          ? false
          : $(input).hasClass('tagify--autocreate');

        // Make tags sortable
        Sortable.create(tagify.DOM.scope, {
          draggable: `.${tagify.settings.classNames.tag}:not(tagify__input)`,
          forceFallback: true,
          onEnd() {
            tagify.updateValueByDOMTags();
          },
        });

        // Event handlers
        tagify.on('input', e => {
          handleAutocomplete(
            e.detail.value,
            tagify.value.map(item => item.entity_id)
          );
        });

        tagify.on('change', () => {
          if (helpers.isTagLimitReached() && cardinality > 1) {
            tagify.settings.enforceWhitelist = false;
            tagify.settings.skipInvalid = false;
          }
        });

        // Handle simulated select behavior
        if (!tagify.settings.dropdown.enabled) {
          $(`.${identifier}`).addClass('tagify-select');

          $(document).on('click', e => {
            const $target = $(e.target);
            if ($target.hasClass('tagify__input') && $target.closest(`.${identifier}`).length) {
              handleAutocomplete(
                '',
                tagify.value.map(item => item.entity_id)
              );
            }
          });
        }
      });
    }
  };
})(jQuery, Drupal, drupalSettings, Sortable);
